/*!
 * SiTCP Remote Bus Control Protocol
 *
 * Copyright (C) 2016
 *		Bee Beans Technologies Co.,Ltd.
 *		All rights reserved.
 */
#include "rbcp.h"
#include <memory.h>
#include <stdio.h>
#include <errno.h>

Rbcp::Rbcp(const std::string &address, int port, int retry, int timeout)
{
	m_id = 0;
	m_retry = retry;
	m_timeout = timeout;
	m_sock = socket(AF_INET, SOCK_DGRAM, 0);
	memset(&m_sitcpAddr, 0, sizeof(m_sitcpAddr));
	m_sitcpAddr.sin_family = AF_INET;
	m_sitcpAddr.sin_port = htons(port);
	m_sitcpAddr.sin_addr.s_addr = inet_addr(address.c_str());
}

int Rbcp::Write(unsigned int address, const char *data, unsigned char length)
{
	int res = 0;
	for (int i = 0; i < m_retry; i++) {
		res = commRBCP(RBCP_HEADER_SIZE, setWriteCmd(address, data, length));
		if (res != Timeout)
			break;
	}
	return res;
}

int Rbcp::setWriteCmd(unsigned int address, const char *data, unsigned char length)
{
	struct rbcp_header header;
	header.type = RBCP_VER;
	header.command = RBCP_WR;
	header.id = getId();
	header.length = length;
	header.address = htonl(address);

	m_buf.resize(length + RBCP_HEADER_SIZE);
	char *buf = m_buf.data();
	memcpy(buf, &header, RBCP_HEADER_SIZE);
	buf += RBCP_HEADER_SIZE;
	memcpy(buf, data, length);
	return header.id;
}

int Rbcp::Read(unsigned int address, char *data, unsigned char length)
{
	int res = 0;
	for (int i = 0; i < m_retry; i++) {
		res = commRBCP(RBCP_HEADER_SIZE + length, setReadCmd(address, length));
		if (res > 0) {
			res -= RBCP_HEADER_SIZE;
			res = std::min(res, (int)length);
			if (data && res > 0) {
				memcpy(data, m_buf.data() + RBCP_HEADER_SIZE, res);
			}
			return res;
		}
		else if (res != Timeout)
			break;
	}
	return res;
}

int Rbcp::setReadCmd(unsigned int address, unsigned char length)
{
	struct rbcp_header header;
	header.type = RBCP_VER;
	header.command = RBCP_RD;
	header.id = getId();
	header.length = length;
	header.address = htonl(address);

	m_buf.resize(RBCP_HEADER_SIZE);
	char *buf = m_buf.data();
	memcpy(buf, &header, RBCP_HEADER_SIZE);
	return header.id;
}

int Rbcp::commRBCP(int recvSize, char id)
{
	sendto(m_sock, m_buf.data(), m_buf.size(), 0, (struct sockaddr *)&m_sitcpAddr, sizeof(m_sitcpAddr));
	fd_set setSelect;
	FD_ZERO(&setSelect);
	FD_SET(m_sock, &setSelect);

	struct timeval timeout;
	timeout.tv_sec	= m_timeout;
	timeout.tv_usec = 0;
 
	if (select(m_sock+1, &setSelect, NULL, NULL, &timeout) == 0) { /* time out */
		return Timeout;
	}
	else if (FD_ISSET(m_sock, &setSelect)) { /* receive packet */
		m_buf.resize(recvSize);
		errno = 0;
		int rcvdBytes = recvfrom(m_sock, m_buf.data(), m_buf.size(), 0, NULL, NULL);
		if (rcvdBytes < RBCP_HEADER_SIZE) {
			//perror("recvfrom()");
			//std::string estr("ERROR: Unable to receive ACK packet");
			int num = errno;
			std::string estr("\nError: recvfrom() - ");
			estr.append(strerror(num));
			throw estr;
		}
		if (m_buf.at(1) & 0x01)
			return BusErr; // Bus Error

		if (m_buf.at(2) == id)
			return rcvdBytes; // on success
	}
	return Timeout; // retry
}
